/* KallistiOS 1.1.7

 conio.c

 (c)2002 Dan Potter

 Adapted from Kosh, (c)2000 Jordan DeLong

*/

#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <kos/sem.h>
#include <dc/maple/keyboard.h>
#include <conio/conio.h>
#include <conio/draw.h>
#include <conio/input.h>

/* the cursor */
conio_cursor_t conio_cursor;

/* the virtual screen */
char conio_virtscr[CONIO_NUM_ROWS][CONIO_NUM_COLS];

/* Our modes */
int conio_ttymode = CONIO_TTY_NONE, conio_inputmode = CONIO_INPUT_NONE;

/* freeze/thaw mutex */
static semaphore_t *ft_mutex;

/* scroll everything up a line */
void conio_scroll() {
	int i;

	memmove(conio_virtscr, conio_virtscr[1], (CONIO_NUM_ROWS - 1) * CONIO_NUM_COLS);
	for (i = 0; i < CONIO_NUM_COLS; i++)
		conio_virtscr[CONIO_NUM_ROWS - 1][i] = ' ';
	conio_cursor.row--;
}

/* move the cursor back, don't scroll (we can't) */
void conio_deadvance_cursor() {
	conio_cursor.col--;
	if (conio_cursor.col < 0) {
		if (conio_cursor.row != 0) {
			conio_cursor.col = CONIO_NUM_COLS - 1;
			conio_cursor.row--;
		} else {
			conio_cursor.col = 0;
		}
	}
}

/* move the cursor ahead, scroll if we need to */
void conio_advance_cursor() {
	conio_cursor.col++;
	if (conio_cursor.col >= CONIO_NUM_COLS) {
		conio_cursor.col = 0;
		conio_cursor.row++;
		if (conio_cursor.row >= CONIO_NUM_ROWS)
			conio_scroll();
	}
}

/* move the cursor */
void conio_gotoxy(int x, int y) {
	conio_cursor.col = x;
	conio_cursor.row = y;
}

/* blocking call for a character */
int conio_getch() {
	int key;

	while ((key = kbd_get_key()) == -1) {}
	return key;
}

/* Check to see if a key has been pressed */
int conio_check_getch() {
	int key;

	key = kbd_get_key();
	return key;
}

/* set current char to ch, w/o advancing the cursor */
void conio_setch(int ch) {
	switch (ch) {
		case '\n':
		case '\r':
			break;
		default:
			conio_virtscr[conio_cursor.row][conio_cursor.col] = ch;
	}
}

/* put a character at the cursor and move the cursor */
void conio_putch(int ch) {
	switch (ch) {
		case '\r':
			conio_cursor.col = 0;
			break;
		case '\n':
			conio_cursor.row++;
			conio_cursor.col = 0;
			if (conio_cursor.row >= CONIO_NUM_ROWS)
				conio_scroll();
			break;
		default:
			conio_virtscr[conio_cursor.row][conio_cursor.col] = ch;
			conio_advance_cursor();
	}
}

/* put a string of characters */
void conio_putstr(char *str) {
	while (*str != '\0') {
		conio_putch(*str++);
	}
}

/* a printfish function */
int conio_printf(const char *fmt, ...) {
	char buff[512];	/* buffer overflow waiting to happen.... I'll add a vsnprintf func later */
	va_list args;
	int i;

	va_start(args, fmt);
	i = vsprintf(buff, fmt, args);

	conio_putstr(buff);

	va_end(args);

	return i;
}

/* clear the screen */
void conio_clear() {
	int row, col;

	/* fill screen with spaces */
	for (row = 0; row < CONIO_NUM_ROWS; row++)
		for (col = 0; col < CONIO_NUM_COLS; col++)
			conio_virtscr[row][col] = ' ';
}

/* conio freeze (for sub-process taking over TA) */
void conio_freeze() {
	sem_wait(ft_mutex);
}

/* conio thaw */
void conio_thaw() {
	sem_signal(ft_mutex);
}

/* ptr to old printf */
/* static int (*oldprintf)(const char *fmt, ...) = NULL; */
static volatile int conio_entered = 0;
static volatile int conio_exit = 0;

/* the drawing/keyboard polling thread */
static void conio_thread(void *param) {
	conio_entered = 1;
	while (!conio_exit) {
		sem_wait(ft_mutex);
		conio_input_frame();
		conio_draw_frame();
		sem_signal(ft_mutex);
	}
	conio_exit = -1;
}

/* initialize the console I/O stuffs */
int conio_init(int ttymode, int inputmode) {
	conio_ttymode = ttymode;
	conio_inputmode = inputmode;

	assert_msg( conio_ttymode == CONIO_TTY_PVR, "Non-PVR TTY modes not supported yet" );
	assert_msg( conio_inputmode == CONIO_INPUT_LINE, "Non-Line input modes not supported yet" );

	conio_draw_init();
	conio_input_init();
	conio_clear();
	conio_gotoxy(0, 0);
	/* conio_svc_init(); */

	ft_mutex = sem_create(1);

	/* replace fundemental printf */
	/* oldprintf = fabi->printf;
	fabi->printf = conio_printf; */

	/* create the conio thread */
	conio_exit = 0;
	if (thd_create(conio_thread, 0) < 0)
		return -1;

	/* Wait for it to actually start */
	while (!conio_entered)
		thd_pass();
	
	return 0;
}

int conio_shutdown() {
	/* shutup our thread */
	conio_exit = 1;

	while (conio_exit != -1)
		;

	/* Delete the sempahore */
	sem_destroy(ft_mutex);

	conio_draw_shutdown();
	conio_input_shutdown();

	conio_ttymode = CONIO_TTY_NONE;
	conio_inputmode = CONIO_INPUT_NONE;
	
	/* put back the old printf */
	/* fabi->printf = oldprintf; */

	return 0;
}
